from dcqc.target import SingleTarget
from dcqc.tests.base_test import ExternalBaseTest, Process


class TiffTag306DateTimeTest(ExternalBaseTest):
    tier = 4
    pass_code = "1"
    target: SingleTarget

    def generate_process(self) -> Process:
        path = self.target.file.stage()
        string_path = self._short_string_path(path, "dcqc-staged-")

        command_args = [
            "tifftools",
            "dump",
            string_path,
            "|",
            "grep",  # pipe the output
            "-a",  # treat input as text
            "-q",  # suppress output
            "'DateTime 306 (0x132) ASCII'",  # match the DateTime 306 tag
        ]
        process = Process(
            container="ghcr.io/sage-bionetworks-workflows/tifftools:latest",
            command_args=command_args,
        )
        return process



"""
    Dummy conftest.py for dcqc.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

from datetime import datetime
from getpass import getuser
from pathlib import Path
from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from dcqc.file import File
from dcqc.suites.suite_abc import SuiteABC, SuiteStatus
from dcqc.target import SingleTarget

CNFPATH = Path(__file__).resolve()
TESTDIR = CNFPATH.parent
DATADIR = TESTDIR / "data"
OUTDIR = TESTDIR / "outputs"

OUTDIR.mkdir(exist_ok=True)

UUID = str(uuid4())
USER = getuser()
UTCTIME = datetime.now().isoformat(" ", "seconds").replace(":", ".")
RUN_ID = f"{USER} - {UTCTIME} - {UUID}"  # Valid characters: [A-Za-z0-9 .+'()_-]


# Track the list of output files to avoid clashes between tests
outputs = set()


@pytest.fixture
def run_id():
    return RUN_ID


@pytest.fixture
def get_data():
    def _get_data(filename: str) -> Path:
        path = DATADIR / filename
        if not path.exists():
            raise ValueError(f"Path ({path}) does not exist.")
        return path

    yield _get_data


@pytest.fixture
def test_files(get_data):
    txt_path = get_data("test.txt")
    jsonld_path = get_data("example.jsonld")
    tiff_path = get_data("circuit.tif")
    fastq1_path = get_data("fastq1.fastq")
    fastq2_path = get_data("fastq2.fastq.gz")
    syn_path = "syn://syn50555279"
    tiff_dirty_datetime_path = get_data("test_image_dirty_datetime.tif")
    good_metadata = {
        "file_type": "txt",
        "md5_checksum": "14758f1afd44c09b7992073ccf00b43d",
    }
    bad_metadata = {
        "file_type": "tiff",
        "md5_checksum": "definitelynottherightmd5checksum",
    }
    jsonld_metadata = {
        "file_type": "JSON-LD",
        "md5_checksum": "56bb5f34da6d6df2ade3ac37e25586b7",
    }
    tiff_metadata = {
        "file_type": "tiff",
        "md5_checksum": "c7b08f6decb5e7572efbe6074926a843",
    }
    fastq_metadata = {"file_type": "fastq"}
    tiff_dirty_datetime_metadata = {
        "file_type": "tiff",
        "md5_checksum": "28a9ee7d0e994d494068ce8d6cda0268",
    }
    test_files = {
        "good": File(txt_path.as_posix(), good_metadata),
        "bad": File(txt_path.as_posix(), bad_metadata),
        "tiff": File(tiff_path.as_posix(), tiff_metadata),
        "fastq1": File(fastq1_path.as_posix(), fastq_metadata),
        "fastq2": File(fastq2_path.as_posix(), fastq_metadata),
        "jsonld": File(jsonld_path.as_posix(), jsonld_metadata),
        "synapse": File(syn_path, good_metadata),
        "tiff_dirty_datetime": File(
            tiff_dirty_datetime_path.as_posix(), tiff_dirty_datetime_metadata
        ),
    }

    # Create an in-memory remote file based on the good file
    remote_file = File(f"mem://{txt_path.name}", good_metadata)
    remote_file.fs.writetext(remote_file.fs_path, txt_path.read_text())
    test_files["remote"] = remote_file

    yield test_files


@pytest.fixture
def test_targets(test_files):
    test_targets = dict()
    for name, file in test_files.items():
        test_targets[name] = SingleTarget(file)
    yield test_targets


@pytest.fixture
def test_suites(test_targets):
    test_suites = dict()
    for name, target in test_targets.items():
        test_suites[name] = SuiteABC.from_target(target)
    yield test_suites


@pytest.fixture
def get_output():
    def _get_output(filename: str) -> Path:
        output = OUTDIR / filename
        if output in outputs:
            message = f"Output ({output}) has already been used in another test."
            raise ValueError(message)
        outputs.add(output)
        return output

    yield _get_output


@pytest.fixture
def mocked_suites_single_targets():
    mock_dict_single = {
        "syn://syn51585496": SuiteStatus.GREEN,
        "syn://syn51585494": SuiteStatus.RED,
        "syn://syn51585495": SuiteStatus.AMBER,
        "syn://syn51585493": SuiteStatus.NONE,
    }
    mocked_suites = []
    for url, status in mock_dict_single.items():
        suite = MagicMock(cls=SuiteABC)
        suite.target.files[0].url = url
        suite.get_status.return_value = status
        mocked_suites.append(suite)
    return mocked_suites


# @pytest.fixture
# def mocked_suites_multi_targets():
#     mock_dict_multi = {
#         "syn://syn51585496": SuiteStatus.GREEN,
#         "syn://syn51585494": SuiteStatus.RED,
#         "syn://syn51585495": SuiteStatus.AMBER,
#     }
#     mocked_suites = []
#     for url, status in mock_dict_multi.items():
#         suite = MagicMock(cls=SuiteABC)
#         suite.target.files[0].url = url
#         suite.get_status.return_value = status
#         mocked_suites.append(suite)
#     return mocked_suites



"""Represent local and remote files and their metadata.

See module dcqc.target for the multi-file target class.

Classes:

    FileType: For collecting file type-specific information.
    File: For bundling file location and metadata as well as
    operations for retrieving file contents.
"""

from __future__ import annotations

import os
from collections.abc import Collection, Mapping
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, ClassVar, Optional
from warnings import warn

from fs.base import FS

from dcqc.mixins import SerializableMixin, SerializedObject
from dcqc.utils import is_url_local, open_parent_fs


@dataclass
class FileType:
    """Bundle information for a given file type."""

    _registry: ClassVar[dict[str, FileType]]
    _registry = dict()

    name: str
    file_extensions: tuple[str, ...]
    edam_iri: Optional[str]

    def __init__(
        self,
        name: str,
        file_extensions: Collection[str],
        edam_iri: Optional[str] = None,
    ):
        """Construct a FileType object.

        Args:
            name: File type name.
            file_extensions: Valid file extensions.
            edam_iri: EDAM format ontology identifier.
        """
        self.name = name
        self.file_extensions = tuple(file_extensions)
        self.edam_iri = edam_iri
        self.register_file_type()

    def register_file_type(self) -> None:
        """Register instantiated file type for later retrieval.

        Raises:
            ValueError: If the file type's name has already
                been registered previously.
        """
        name = self.name.lower()
        if name in self._registry:
            message = f"File type ({name}) is already registered ({self._registry})."
            raise ValueError(message)
        self._registry[name] = self

    @classmethod
    def list_file_types(cls) -> list[FileType]:
        """Retrieve all available file type objects.

        Returns:
            The full list of file type objects.
        """
        return list(cls._registry.values())

    @classmethod
    def get_file_type(cls, file_type: str) -> FileType:
        """Retrieve file type object based on its name.

        Args:
            file_type: File type name.

        Raises:
            ValueError: If the given file type name has
                not been registered previously.

        Returns:
            The file type object with the given name.
        """
        file_type = file_type.lower()
        if file_type not in cls._registry:
            types = list(cls._registry)
            message = f"File type ({file_type}) not among available options ({types})."
            raise ValueError(message)
        return cls._registry[file_type]


# TODO: These file types could be moved to an external file
# Instantiated file types are automatically tracked by the FileType class
FileType("*", (), "format_1915")  # To represent all file types
FileType("TXT", (".txt",), "format_1964")
FileType("JSON", (".json",), "format_3464")
FileType("JSON-LD", (".jsonld",), "format_3749")
FileType("TIFF", (".tif", ".tiff"), "format_3591")
FileType("OME-TIFF", (".ome.tif", ".ome.tiff"), "format_3727")
FileType("TSV", (".tsv"), "format_3475")
FileType("BAM", (".bam"), "format_2572")
FileType("FASTQ", (".fastq", ".fastq.gz", ".fq", ".fq.gz"), "format_1930")


# TODO: Leverage post-init function in dataclasses
@dataclass
class File(SerializableMixin):
    """Construct a File object.

    Args:
        url: Local or remote location of a file.
        metadata: File metadata.
        relative_to: Used to update any local URLs if they
            are relative to a directory other than the
            current work directory (default).
    """

    _serialized_properties = ["name", "local_path"]

    url: str
    metadata: dict[str, Any]
    type: str

    def __init__(
        self,
        url: str,
        metadata: Optional[Mapping[str, Any]] = None,
        relative_to: Optional[Path] = None,
        local_path: Optional[Path] = None,
    ):
        self.url = self._relativize_url(url, relative_to)
        metadata = metadata or dict()
        self.metadata = dict(metadata)
        self.type = self._pop_file_type()

        self._fs: Optional[FS]
        self._fs = None
        self._fs_path: Optional[str]
        self._fs_path = None
        self._name: Optional[str]
        self._name = None
        self._local_path: Optional[Path]
        self._local_path = local_path

    def __hash__(self):
        return hash((self.url, self.type, tuple(self.metadata.items())))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def _relativize_url(self, url: str, relative_to: Optional[Path]) -> str:
        """Update local URLs if relative to a directory other than CWD.

        Args:
            url: Local or remote location of a file.
            relative_to: Used to update any local URLs if they
                are relative to a directory other than the
                current work directory (default).

        Returns:
            The relativized URL.
        """
        if self.is_url_local(url):
            relative_to = relative_to or Path.cwd()
            scheme, separator, resource = url.rpartition("://")
            path = Path(resource)
            if not path.is_absolute():
                resource = os.path.relpath(relative_to / resource)
            url = f"{scheme}{separator}{resource}"
        elif not self.is_url_local(url) and relative_to is not None:
            message = f"URL ({url}) is remote. Ignoring relative_to ({relative_to})."
            warn(message)
        return url

    def _pop_file_type(self) -> str:
        """Extract and remove file type from metadata.

        This function defaults to the generic file type
        ("*") if the key is absent from the metadata.

        Returns:
            The name of the file type in the metadata.
        """
        file_type = self.metadata.pop("file_type", "*")
        return file_type

    def _init_fs(self) -> tuple[FS, str]:
        """Initialize file system to access URL.

        All queries with this file system should use
        `self._fs_path` as the path, not `self.url`.

        Returns:
            A file system + basename pair.
        """
        fs, fs_path = open_parent_fs(self.url)
        self._fs_path = fs_path
        self._fs = fs
        return fs, fs_path

    @property
    def local_path(self) -> Path:
        """Retrieve the path to a local copy if available.

        Raises:
            FileNotFoundError: If a remote file has not been
                staged yet and thus has no local copy.

        Returns:
            The path to the local copy or `None` if unavailable.
        """
        if self._local_path is None and self.is_url_local():
            _local_path = self.fs.getsyspath(self.fs_path)
            self._local_path = Path(_local_path)
        if self._local_path is None:
            message = "Local path is unavailable. Use stage() to create a local copy."
            raise FileNotFoundError(message)
        return self._local_path

    @property
    def fs(self) -> FS:
        """The file system that can access the URL."""
        fs = self._fs
        if fs is None:
            fs, _ = self._init_fs()
        return fs

    @property
    def fs_path(self) -> str:
        """The path that can be used with the file system."""
        fs_path = self._fs_path
        if fs_path is None:
            _, fs_path = self._init_fs()
        return fs_path

    @property
    def name(self) -> str:
        """The file name according to the file system."""
        if self._name is None:
            info = self.fs.getinfo(self.fs_path)
            self._name = info.name
        return self._name

    def get_file_type(self) -> FileType:
        """Retrieve the relevant file type object.

        Returns:
            FileType: File type object
        """
        return FileType.get_file_type(self.type)

    def get_metadata(self, key: str) -> Any:
        """Retrieve file metadata using a key.

        Args:
            key: Metadata key name.

        Raises:
            KeyError: If the metadata key doesn't exist.

        Returns:
            The metadata value associated with the given key.
        """
        if key not in self.metadata:
            url = self.url
            md = self.metadata
            message = f"File ({url}) does not have '{key}' in its metadata ({md})."
            raise KeyError(message)
        return self.metadata[key]

    def is_url_local(self, url: Optional[str] = None) -> bool:
        """Check whether a URL refers to a local location.

        Args:
            url: Local or remote location of a file.
                Defaults to URL associated with file.

        Returns:
            Whether the URL refers to a local location.
        """
        url = url or self.url
        return is_url_local(url)

    def is_file_local(self) -> bool:
        """Check if the file (or a copy) is available locally.

        Unlike :func:`~dcqc.file.File.is_url_local`, this method
        considers if a locally staged copy is available regardless
        of whether the URL is local or remote.

        To retrieve the location of the local copy, you can use
        :attr:`~dcqc.file.File.local_path

        Returns:
            Whether the file has a copy available locally.
        """
        return self._local_path is not None

    def stage(
        self,
        destination: Optional[Path] = None,
        overwrite: bool = False,
    ) -> Path:
        """Create local copy of local or remote file.

        A destination is not required for remote files; it
        defaults to a temporary directory.
        Local files aren't moved if a destination is omitted.

        Args:
            destination: File or folder where to store the file.
                Defaults to None.
            overwrite: Whether to ignore existing file at the
                target destination. Defaults to False.

        Raises:
            ValueError: If the parent directory of the
                destination does not exist.
            FileExistsError: If the destination file already
                exists and ``overwrite`` was not enabled.

        Returns:
            The path of the local copy.
        """
        if not destination:
            if self._local_path is not None:
                return self._local_path
            else:
                # TODO: This prefix is used by nf-dcqc to easily find the staged file.
                #       It might be worth using a DCQCTMPDIR to avoid hard-coding this.
                destination_str = mkdtemp(prefix="dcqc-staged-")
                destination = Path(destination_str)

        # By this point, destination is defined (not None)
        if destination.is_dir():
            destination = destination / self.name

        if not destination.parent.exists():
            dest = str(destination)
            message = f"Parent folder of destination ({dest}) does not exist."
            raise ValueError(message)

        if destination.exists() and not overwrite:
            dest = str(destination)
            message = f"Destination ({dest}) already exists. Enable overwrite."
            raise FileExistsError(message)

        # By this point, the file either doesn't exist or overwrite is enabled
        destination.unlink(missing_ok=True)

        if self._local_path and self.is_url_local():
            destination.symlink_to(self._local_path.resolve())
        else:
            with destination.open("wb") as dest_file:
                self.fs.download(self.fs_path, dest_file)

        self._local_path = destination
        return destination

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> File:
        """Deserialize a dictionary into a file.

        Args:
            dictionary: A serialized file object.

        Returns:
            The reconstructed file object.
        """
        dictionary = deepcopy(dictionary)

        file_type = dictionary.pop("type")
        dictionary["metadata"]["file_type"] = file_type

        if dictionary["local_path"] is not None:
            dictionary["local_path"] = Path(dictionary["local_path"])

        # Ignore serialized name since it's a dynamically-computed property
        dictionary.pop("name", None)

        return cls(**dictionary)



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from dcqc import tests
from dcqc.target import PairedTarget, SingleTarget
from dcqc.tests import BaseTest, ExternalTestMixin, Process, TestStatus


def test_that_the_file_extension_test_works_on_correct_files(test_targets):
    target = test_targets["good"]
    test = tests.FileExtensionTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_the_file_extension_test_works_on_correct_remote_file(test_targets):
    target = test_targets["remote"]
    test = tests.FileExtensionTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_a_tiff_file_with_good_extensions_is_passed(test_targets):
    target = test_targets["tiff"]
    test = tests.FileExtensionTest(target)
    assert test.get_status() == TestStatus.PASS


def test_that_the_file_extension_test_works_on_incorrect_files(test_targets):
    target = test_targets["bad"]
    test = tests.FileExtensionTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.FAIL


def test_that_the_md5_checksum_test_works_on_a_correct_file(test_targets):
    target = test_targets["good"]
    test = tests.Md5ChecksumTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_the_md5_checksum_test_works_on_incorrect_files(test_targets):
    target = test_targets["bad"]
    test = tests.Md5ChecksumTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.FAIL


def test_that_the_json_load_test_works_on_a_correct_file(test_targets):
    target = test_targets["jsonld"]
    test = tests.JsonLoadTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_the_json_load_test_works_on_incorrect_files(test_targets):
    target = test_targets["good"]
    test = tests.JsonLoadTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.FAIL


def test_that_the_jsonld_load_test_works_on_a_correct_file(test_targets):
    target = test_targets["jsonld"]
    test = tests.JsonLdLoadTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_the_jsonld_load_test_works_on_incorrect_files(test_targets):
    target = test_targets["good"]
    test = tests.JsonLdLoadTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.FAIL


def test_that_all_external_tests_inherit_from_the_mixin_first():
    tests = BaseTest.list_subclasses()
    for test in tests:
        if issubclass(test, ExternalTestMixin):
            mro = test.__mro__
            mixin_index = mro.index(ExternalTestMixin)
            abc_index = mro.index(BaseTest)
            assert mixin_index < abc_index


def test_that_the_libtiff_info_test_correctly_interprets_exit_code_0_and_1(
    test_files, mocker
):
    # 0 is pass, 1 is fail
    tiff_file = test_files["tiff"]
    target = SingleTarget(tiff_file)
    with TemporaryDirectory() as tmp_dir:
        path_0 = Path(tmp_dir, "code_0.txt")
        path_1 = Path(tmp_dir, "code_1.txt")
        path_0.write_text("0")
        path_1.write_text("1")
        pass_outputs = {"std_out": path_1, "std_err": path_1, "exit_code": path_0}
        fail_outputs = {"std_out": path_0, "std_err": path_0, "exit_code": path_1}

        test = tests.LibTiffInfoTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=pass_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.PASS

        test = tests.LibTiffInfoTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=fail_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.FAIL


def test_that_the_libtiff_info_test_command_is_produced(test_targets):
    target = test_targets["tiff"]
    test = tests.LibTiffInfoTest(target)
    process = test.generate_process()
    assert "tiffinfo" in process.command


def test_that_the_bioformats_info_test_command_is_produced(test_targets):
    target = test_targets["tiff"]
    test = tests.BioFormatsInfoTest(target)
    process = test.generate_process()
    assert "showinf" in process.command


def test_that_the_ome_xml_schema_test_command_is_produced(test_targets):
    target = test_targets["tiff"]
    test = tests.OmeXmlSchemaTest(target)
    process = test.generate_process()
    assert "xmlvalid" in process.command


def test_that_the_md5_checksum_test_can_be_retrieved_by_name():
    test = BaseTest.get_subclass_by_name("Md5ChecksumTest")
    assert test is tests.Md5ChecksumTest


def test_for_an_error_when_retrieving_a_random_test_by_name():
    with pytest.raises(ValueError):
        BaseTest.get_subclass_by_name("FooBar")


def test_that_process_output_files_can_be_found(get_data):
    std_out = get_data("tiffinfo/std_out.txt")
    std_err = get_data("tiffinfo/std_err.txt")
    exit_code = get_data("tiffinfo/exit_code.txt")
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        (tmp_path / std_out.name).symlink_to(std_out)
        (tmp_path / std_err.name).symlink_to(std_err)

        with pytest.raises(FileNotFoundError):
            ExternalTestMixin._find_process_outputs(tmp_path)

        (tmp_path / exit_code.name).symlink_to(exit_code)
        ExternalTestMixin._find_process_outputs(tmp_path)


def test_for_error_when_importing_unavailable_module(test_targets):
    target = test_targets["good"]
    test = tests.FileExtensionTest(target)
    with pytest.raises(ModuleNotFoundError):
        test.import_module("foobar")


def test_that_an_existing_module_can_be_imported(test_targets):
    target = test_targets["good"]
    test = tests.FileExtensionTest(target)
    imported = test.import_module("pytest")
    assert imported is pytest


def test_that_a_process_can_be_serialized_and_deserialized():
    process = Process("foo:bar", ["python"])
    process_dict = process.to_dict()
    process_from_dict = Process.from_dict(process_dict)
    assert process_dict == process_from_dict.to_dict()


def test_that_the_grep_date_test_correctly_interprets_exit_code_0_and_1(
    test_files, mocker
):
    # 1 is pass, 0 is fail
    tiff_file = test_files["tiff"]
    target = SingleTarget(tiff_file)
    with TemporaryDirectory() as tmp_dir:
        path_0 = Path(tmp_dir, "code_0.txt")
        path_1 = Path(tmp_dir, "code_1.txt")
        path_0.write_text("0")
        path_1.write_text("1")
        fail_outputs = {"std_out": path_1, "std_err": path_1, "exit_code": path_0}
        pass_outputs = {"std_out": path_0, "std_err": path_0, "exit_code": path_1}

        test = tests.GrepDateTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=pass_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.PASS

        test = tests.GrepDateTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=fail_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.FAIL


def test_that_the_grep_date_test_command_is_produced(test_targets):
    target = test_targets["tiff"]
    test = tests.GrepDateTest(target)
    process = test.generate_process()
    assert "grep" in process.command


def test_that_the_tifftag306datetimetest_command_is_produced(test_targets):
    target = test_targets["tiff"]
    test = tests.TiffTag306DateTimeTest(target)
    process = test.generate_process()
    assert "grep" in process.command


def test_that_the_tifftag306datetimetest_correctly_interprets_exit_code_0_and_1(
    test_files, mocker
):
    # 1 is pass, 0 is fail
    tiff_file = test_files["tiff"]
    target = SingleTarget(tiff_file)
    with TemporaryDirectory() as tmp_dir:
        path_0 = Path(tmp_dir, "code_0.txt")
        path_1 = Path(tmp_dir, "code_1.txt")
        path_0.write_text("0")
        path_1.write_text("1")
        fail_outputs = {"std_out": path_1, "std_err": path_1, "exit_code": path_0}
        pass_outputs = {"std_out": path_0, "std_err": path_0, "exit_code": path_1}

        test = tests.TiffTag306DateTimeTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=pass_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.PASS

        test = tests.TiffTag306DateTimeTest(target)
        mocker.patch.object(test, "_find_process_outputs", return_value=fail_outputs)
        test_status = test.get_status()
        assert test_status == TestStatus.FAIL


def test_that_paired_fastq_parity_test_correctly_passes_identical_fastq_files(
    test_files,
):
    fastq1 = test_files["fastq1"]
    target = PairedTarget([fastq1, fastq1])
    test = tests.PairedFastqParityTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_paired_fastq_parity_test_correctly_fails_different_fastq_files(
    test_files,
):
    fastq1 = test_files["fastq1"]
    fastq2 = test_files["fastq2"]
    target = PairedTarget([fastq1, fastq2])
    test = tests.PairedFastqParityTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.FAIL


def test_that_paired_fastq_parity_test_correctly_handles_compressed_fastq_files(
    test_files,
):
    fastq2 = test_files["fastq2"]
    target = PairedTarget([fastq2, fastq2])
    test = tests.PairedFastqParityTest(target)
    test_status = test.get_status()
    assert test_status == TestStatus.PASS


def test_that_short_string_path_correctly_shortens_file_paths():
    substring = "test-substring"
    long_path = f"path/needs/to/be/shortened/{substring}/file.txt"
    expected_short_path = f"'{substring}/file.txt'"
    short_path = ExternalTestMixin._short_string_path(Path(long_path), substring)
    assert short_path == expected_short_path


def test_that_short_string_path_raises_valueerror_if_substring_not_in_path():
    substring = "test-substring"
    long_path = "path/needs/to/be/shortened/fail/file.txt"
    with pytest.raises(ValueError):
        ExternalTestMixin._short_string_path(Path(long_path), substring)



import json
import sys
from csv import DictWriter
from pathlib import Path
from typing import List

from typer import Argument, Exit, Option, Typer

from dcqc import __version__
from dcqc.file import File, FileType
from dcqc.parsers import CsvParser, JsonParser
from dcqc.reports import JsonReport
from dcqc.suites.suite_abc import SuiteABC
from dcqc.target import SingleTarget
from dcqc.tests.base_test import BaseTest, ExternalTestMixin
from dcqc.updaters import CsvUpdater

# Make commands optional to allow for `dcqc --version`
app = Typer(invoke_without_command=True)


# Common arguments
# Distinguishing between path and path/url arguments until I/O is consistent
input_path_arg = Argument(..., help="Input file")
input_path_list_arg = Argument(..., help="List of input files")
output_arg = Argument(..., help="Path or (remote) URL for output file")
output_dir_arg = Argument(..., help="Directory path or (remote) URL for output files")
output_path_arg = Argument(..., help="Path for output file")
output_dir_path_arg = Argument(..., help="Directory path for output files")

# Common options
overwrite_opt = Option(False, "--overwrite", "-f", help="Ignore existing files")
required_tests_opt = Option(None, "--required-tests", "-r", help="Required tests")
skipped_tests_opt = Option(None, "--skipped-tests", "-s", help="Skipped tests")
file_type_opt = Option(..., "--file-type", "-t", help="File type")
metadata_opt = Option("{}", "--metadata", "-m", help="File metadata")


@app.callback()
def main(version: bool = False):
    """DCQC Python Package"""
    if version:
        print(f"DCQC Python Package Version: {__version__}")
        raise Exit()


@app.command()
def create_targets(
    input_csv: Path = input_path_arg,
    output_dir: Path = output_dir_path_arg,
    overwrite: bool = overwrite_opt,
):
    """Create target JSON files from a targets CSV file"""
    output_dir.mkdir(parents=True, exist_ok=True)

    parser = CsvParser(input_csv)
    targets = parser.create_targets()

    # Naming the targets by index to ensure no clashes
    named_targets = {f"target-{target.id}.json": target for target in targets}

    report = JsonReport()
    report.save_many(named_targets, output_dir.as_posix(), overwrite)


@app.command()
def create_tests(
    input_json: Path = input_path_arg,
    output_dir: Path = output_dir_path_arg,
    required_tests: List[str] = required_tests_opt,
    skipped_tests: List[str] = skipped_tests_opt,
    overwrite: bool = overwrite_opt,
):
    """Create test JSON files from a target JSON file"""
    output_dir.mkdir(parents=True, exist_ok=True)

    # Interpret empty lists from CLI as None (to auto-generate values)
    required_tests_maybe = required_tests if required_tests else None
    skipped_tests_maybe = skipped_tests if skipped_tests else None

    target = JsonParser.parse_object(input_json, SingleTarget)
    suite = SuiteABC.from_target(target, required_tests_maybe, skipped_tests_maybe)

    report = JsonReport()
    for test in suite.tests:
        output_path = output_dir / f"{input_json.stem}.{test.type}.json"
        output_url = output_path.as_posix()
        report.save(test, output_url, overwrite)


@app.command()
def create_process(
    input_json: Path = input_path_arg,
    output_json: Path = output_path_arg,
    overwrite: bool = overwrite_opt,
):
    """Create external process JSON file from a test JSON file"""
    output_json.parent.mkdir(parents=True, exist_ok=True)

    test = JsonParser.parse_object(input_json, ExternalTestMixin)
    process = test.generate_process()
    output_url = output_json.as_posix()

    report = JsonReport()
    report.save(process, output_url, overwrite)


@app.command()
def compute_test(
    input_json: Path = input_path_arg,
    output_json: Path = output_path_arg,
    overwrite: bool = overwrite_opt,
):
    """Compute the test status from a test JSON file"""
    output_json.parent.mkdir(parents=True, exist_ok=True)

    test = JsonParser.parse_object(input_json, BaseTest)
    test.get_status()
    output_url = output_json.as_posix()

    report = JsonReport()
    report.save(test, output_url, overwrite)


@app.command()
def create_suite(
    output_json: str = output_arg,
    input_jsons: List[Path] = input_path_list_arg,
    required_tests: List[str] = required_tests_opt,
    skipped_tests: List[str] = skipped_tests_opt,
    overwrite: bool = overwrite_opt,
):
    """Create a suite from a set of test JSON files sharing the same target"""
    # Interpret empty lists from CLI as None (to auto-generate values)
    required_tests_maybe = required_tests if required_tests else None
    skipped_tests_maybe = skipped_tests if skipped_tests else None

    tests = [JsonParser.parse_object(test_json, BaseTest) for test_json in input_jsons]
    suite = SuiteABC.from_tests(tests, required_tests_maybe, skipped_tests_maybe)
    report = JsonReport()
    report.save(suite, output_json, overwrite)


@app.command()
def combine_suites(
    output_json: str = output_arg,
    input_jsons: List[Path] = input_path_list_arg,
    overwrite: bool = overwrite_opt,
):
    """Combine several suite JSON files into a single JSON report"""
    suites = [JsonParser.parse_object(json_, SuiteABC) for json_ in input_jsons]
    report = JsonReport()
    report.save(suites, output_json, overwrite)


@app.command()
def list_tests():
    """List the tests available for each file type"""
    test_classes_by_file_type = SuiteABC.list_test_classes_by_file_type()

    rows = list()
    for file_type_name, test_classes in test_classes_by_file_type.items():
        file_type = FileType.get_file_type(file_type_name)
        for test_cls in test_classes:
            test_dict = {
                "file_type": file_type_name,
                "edam_iri": file_type.edam_iri,
                "test_name": test_cls.__name__,
                "test_tier": test_cls.tier,
                "test_type": "external" if test_cls.is_external_test else "internal",
            }
            rows.append(test_dict)

    fieldnames = list(rows[0])
    writer = DictWriter(sys.stdout, fieldnames)
    writer.writeheader()
    writer.writerows(rows)


@app.command()
def qc_file(
    input_file: Path = input_path_arg,
    file_type: str = file_type_opt,
    metadata: str = metadata_opt,
    required_tests: List[str] = required_tests_opt,
    skipped_tests: List[str] = skipped_tests_opt,
):
    """Run QC tests on a single file (external tests are skipped)"""
    # Interpret empty lists from CLI as None (to auto-generate values)
    required_tests_maybe = required_tests if required_tests else None

    # Prepare target
    file_metadata = json.loads(metadata)
    file_metadata["file_type"] = file_type
    file = File(input_file.as_posix(), file_metadata)
    target = SingleTarget(file)

    # Prepare suite (skip all external tests)
    suite = SuiteABC.from_target(target, required_tests_maybe, skipped_tests)
    external_tests = [test.type for test in suite.tests if test.is_external_test]
    skipped_tests += external_tests
    suite = SuiteABC.from_target(target, required_tests_maybe, skipped_tests)

    # Output QC report on stdout
    report = JsonReport()
    suite_json = report.generate(suite)
    json.dump(suite_json, sys.stdout, indent=2)


@app.command()
def update_csv(
    suites_file: Path = input_path_arg,
    input_file: Path = input_path_arg,
    output_file: Path = output_path_arg,
):
    """Update input CSV file with dcqc_status column"""
    suites = JsonParser.parse_objects(suites_file, SuiteABC)
    updater = CsvUpdater(input_file, output_file)
    updater.update(suites)



from dcqc.target import SingleTarget
from dcqc.tests.base_test import InternalBaseTest, TestStatus


class FileExtensionTest(InternalBaseTest):
    tier = 1
    target: SingleTarget

    def compute_status(self) -> TestStatus:
        status = TestStatus.PASS
        for file in self.target.files:
            file_type = file.get_file_type()
            file_extensions = file_type.file_extensions
            if not file.name.endswith(file_extensions):
                status = TestStatus.FAIL
                break
        return status



import csv
import json
from collections.abc import Collection, Iterator
from pathlib import Path
from typing import Any, Optional, Type, TypeVar, cast

from dcqc.file import File, FileType
from dcqc.mixins import SerializableMixin
from dcqc.suites.suite_abc import SuiteABC
from dcqc.target import BaseTarget, SingleTarget
from dcqc.tests.base_test import BaseTest

# For context on TypeVar, check out this GitHub PR comment:
# https://github.com/Sage-Bionetworks-Workflows/py-dcqc/pull/8#discussion_r1087141497
T = TypeVar("T", bound=SerializableMixin)


# TODO: Add support for URLs instead of paths
# TODO: Add support for a `unique_id` column
class CsvParser:
    path: Path
    stage_files: bool

    def __init__(self, path: Path, stage_files: bool = False):
        self.path = path
        self.stage_files = stage_files

    def list_rows(self) -> Iterator[tuple[int, dict]]:
        with self.path.open(newline="") as file:
            reader = csv.DictReader(file)
            for index, row in enumerate(reader, start=1):
                yield index, row

    def _row_to_file(self, row: dict[str, str]) -> File:
        url = row.pop("url")
        file = File(url, row, relative_to=self.path.parent)
        return file

    def create_files(self) -> Iterator[tuple[int, File]]:
        for index, row in self.list_rows():
            file = self._row_to_file(row)
            if not file.is_file_local() and self.stage_files:
                destination = self.path.parent / "staged_files" / f"index_{index}"
                destination.mkdir(parents=True, exist_ok=True)
                file.stage(destination, overwrite=True)
            yield index, file

    def create_targets(self) -> Iterator[SingleTarget]:
        for index, file in self.create_files():
            yield SingleTarget(file, id=f"{index:04}")

    def create_suites(
        self,
        required_tests: Optional[Collection[str]] = None,
        skipped_tests: Optional[Collection[str]] = None,
    ) -> Iterator[SuiteABC]:
        for target in self.create_targets():
            yield SuiteABC.from_target(target, required_tests, skipped_tests)


class JsonParser:
    path: Path

    def __init__(self, path: Path):
        self.path = path

    def load_json(self) -> Any:
        with self.path.open("r") as infile:
            contents = json.load(infile)
        return contents

    def check_expected_cls(self, instance: Any, expected_cls: Type[T]) -> T:
        if not isinstance(instance, expected_cls):
            cls_name = expected_cls.__name__
            message = f"JSON file ({self.path!s}) is not expected type ({cls_name})."
            raise ValueError(message)
        instance = cast(T, instance)
        return instance

    @classmethod
    def get_class(cls, cls_name: str) -> Type[SerializableMixin]:
        test_classes = BaseTest.list_subclasses()
        test_cls_map = {cls.__name__: cls for cls in test_classes}

        suite_classes = SuiteABC.list_subclasses()
        suite_cls_map = {cls.__name__: cls for cls in suite_classes}

        target_classes = BaseTarget.list_subclasses()
        target_cls_map = {cls.__name__: cls for cls in target_classes}

        file_types = FileType.list_file_types()
        file_types_names = {ft.name.lower() for ft in file_types}

        if cls_name in target_cls_map:
            return target_cls_map[cls_name]
        elif cls_name in test_cls_map:
            return test_cls_map[cls_name]
        elif cls_name in suite_cls_map:
            return suite_cls_map[cls_name]
        elif cls_name.lower() in file_types_names:
            return File
        else:
            message = f"Type ({cls_name}) is not recognized."
            raise ValueError(message)

    @classmethod
    def from_dict(cls, dictionary) -> SerializableMixin:
        if "type" not in dictionary:
            message = f"Cannot parse JSON object due to missing type ({dictionary})."
            raise ValueError(message)
        type_name = dictionary["type"]
        type_class = cls.get_class(type_name)
        object_ = type_class.from_dict(dictionary)
        return object_

    @classmethod
    def parse_object(cls, path: Path, expected_cls: Type[T]) -> T:
        parser = cls(path)
        contents = parser.load_json()

        if isinstance(contents, list):
            message = f"JSON file ({parser.path}) contains a list of objects."
            raise ValueError(message)

        object_ = cls.from_dict(contents)
        expected = parser.check_expected_cls(object_, expected_cls)
        return expected

    @classmethod
    def parse_objects(cls, path: Path, expected_cls: Type[T]) -> list[T]:
        parser = cls(path)
        contents = parser.load_json()

        if not isinstance(contents, list):
            message = f"JSON file ({parser.path}) does not contain a list of objects."
            raise ValueError(message)

        objects = [cls.from_dict(dictionary) for dictionary in contents]
        expected = [parser.check_expected_cls(obj, expected_cls) for obj in objects]
        return expected



import re

from fs import open_fs
from fs.base import FS

LOCAL_URL_REGEX = re.compile(r"((file|osfs)://)?/?[^:]+")


def is_url_local(url: str) -> bool:
    """Check whether a URL refers to a local location.

    Args:
        url: Local or remote location of a file.

    Returns:
        Whether the URL refers to a local location.
    """
    return LOCAL_URL_REGEX.fullmatch(url) is not None


def open_parent_fs(url: str) -> tuple[FS, str]:
    # Split off prefix to avoid issues with `rpartition("/")`
    scheme, separator, resource = url.rpartition("://")
    if separator == "":
        prefix = "osfs://"
    else:
        prefix = scheme + separator

    # Retrieve the "top-most" parent folder for the FS root
    # to ensure that it exists for the FS to be constructed.
    # The remainder of the string can be used as the FS path
    fs_root, _, path = resource.partition("/")

    # Handle the case when the path starts with "/"
    if fs_root == "":
        fs_root = "/"

    # Handle the case when there is no "/" in the path
    if path == "":
        path = fs_root
        fs_root = ""

    fs_url = prefix + fs_root
    fs = open_fs(fs_url)
    return fs, path



from dcqc.target import SingleTarget
from dcqc.tests.base_test import ExternalBaseTest, Process


class LibTiffInfoTest(ExternalBaseTest):
    tier = 2
    pass_code = "0"
    target: SingleTarget

    def generate_process(self) -> Process:
        path = self.target.file.stage()
        string_path = self._short_string_path(path, "dcqc-staged-")

        command_args = [
            "tiffinfo",
            string_path,
        ]
        process = Process(
            container="quay.io/sagebionetworks/libtiff:2.0",
            command_args=command_args,
        )
        return process



from dcqc import tests
from dcqc.file import FileType
from dcqc.suites.suite_abc import SuiteABC


# TODO: Consider moving the filetype-test association logic
#       to the file types
class FileSuite(SuiteABC):
    file_type = FileType.get_file_type("*")
    add_tests = (tests.FileExtensionTest, tests.Md5ChecksumTest)


class JsonSuite(FileSuite):
    file_type = FileType.get_file_type("JSON")
    add_tests = (tests.JsonLoadTest,)


class JsonLdSuite(JsonSuite):
    file_type = FileType.get_file_type("JSON-LD")
    add_tests = (tests.JsonLdLoadTest,)


class TiffSuite(FileSuite):
    file_type = FileType.get_file_type("TIFF")
    add_tests = (
        tests.LibTiffInfoTest,
        tests.GrepDateTest,
        tests.TiffTag306DateTimeTest,
    )


class OmeTiffSuite(TiffSuite):
    file_type = FileType.get_file_type("OME-TIFF")
    add_tests = (tests.OmeXmlSchemaTest, tests.BioFormatsInfoTest)


class TSVSuite(FileSuite):
    file_type = FileType.get_file_type("TSV")


class BAMSuite(FileSuite):
    file_type = FileType.get_file_type("BAM")


class FastqSuite(FileSuite):
    file_type = FileType.get_file_type("FASTQ")
    add_tests = (tests.PairedFastqParityTest,)



"""Top-level dcqc module."""

# isort: skip_file

from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

import logging

# Import suites to ensure that they are defined and thus discoverable
# It is located here to avoid a circular import
from dcqc import tests
from dcqc.suites import suite_abc
from dcqc.suites import suites

# Set default logging handler to avoid "No handler found" warnings
logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.captureWarnings(True)



from dcqc.tests.base_test import BaseTest, ExternalTestMixin, Process, TestStatus
from dcqc.tests.bioformats_info_test import BioFormatsInfoTest
from dcqc.tests.file_extension_test import FileExtensionTest
from dcqc.tests.grep_date_test import GrepDateTest
from dcqc.tests.json_load_test import JsonLoadTest
from dcqc.tests.jsonld_load_test import JsonLdLoadTest
from dcqc.tests.libtiff_info_test import LibTiffInfoTest
from dcqc.tests.md5_checksum_test import Md5ChecksumTest
from dcqc.tests.ome_xml_schema_test import OmeXmlSchemaTest
from dcqc.tests.paired_fastq_parity_test import PairedFastqParityTest
from dcqc.tests.tiff_tag_306_date_time_test import TiffTag306DateTimeTest



from __future__ import annotations

import shlex
from abc import ABC, abstractmethod
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import InitVar, dataclass
from enum import Enum
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import ClassVar, Generic, Optional, TypeVar

from dcqc.mixins import SerializableMixin, SerializedObject, SubclassRegistryMixin
from dcqc.target import BaseTarget

Target = TypeVar("Target", bound=BaseTarget)


class TestStatus(Enum):
    NONE = "pending"
    FAIL = "failed"
    PASS = "passed"
    SKIP = "skipped"


# TODO: Look into the @typing.final decorator
class BaseTest(SerializableMixin, SubclassRegistryMixin, ABC, Generic[Target]):
    """Abstract base class for QC tests.

    Args:
        target: Single- or multi-file target.
        skip: Whether to skip this test, resulting
            in ``TestStatus.SKIP`` as status.
            Defaults to False.

    Raises:
        ValueError: If the test expects a single file and
            the given target features multiple files.
    """

    # Class attributes
    tier: ClassVar[int]
    is_external_test: ClassVar[bool]
    is_external_test = False

    # Instance attributes
    type: str
    target: Target

    def __init__(self, target: Target, skip: bool = False):
        self.type = self.__class__.__name__
        self.target = target
        self._status = TestStatus.SKIP if skip else TestStatus.NONE

    def skip(self):
        """Force the test to be skipped."""
        self._status = TestStatus.SKIP

    def get_status(self, compute_ok: bool = True) -> TestStatus:
        """Compute (if applicable) and return the test status."""
        if self._status == TestStatus.NONE and compute_ok:
            self._status = self.compute_status()
        return self._status

    @abstractmethod
    def compute_status(self) -> TestStatus:
        """Compute the status of the test."""

    def to_dict(self) -> SerializedObject:
        test_dict = {
            "type": self.type,
            "tier": self.tier,
            "is_external_test": self.is_external_test,
            "status": self._status.value,
            "target": self.target.to_dict(),
        }
        return test_dict

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> BaseTest:
        """Deserialize a dictionary into a test.

        Args:
            dictionary: A serialized test object.

        Returns:
            The reconstructed test object.
        """
        test_cls_name = dictionary.pop("type")
        test_cls = BaseTest.get_subclass_by_name(test_cls_name)

        target_dict = dictionary["target"]
        target = BaseTarget.from_dict(target_dict)

        test = test_cls(target)

        status = TestStatus(dictionary["status"])
        test._status = status

        return test

    def import_module(self, name: str) -> ModuleType:
        try:
            module = import_module(name)
        except ModuleNotFoundError:
            message = (
                f"{self.type} cannot be computed without the '{name}' package. ",
                "Re-install `dcqc` with the `all` extra: pip install dcqc[all].",
            )
            raise ModuleNotFoundError(message)
        return module

    @classmethod
    def get_base_class(cls):
        """Retrieve base class."""
        return BaseTest


@dataclass
class Process(SerializableMixin):
    container: str
    command_args: InitVar[Sequence[str]]
    cpus: int = 1
    memory: int = 2  # In GB

    _serialized_properties = ["command"]

    def __post_init__(self, command_args: Sequence[str]):
        self._command_args = command_args

    @property
    def command(self) -> str:
        args_strings = [str(arg) for arg in self._command_args]
        return " ".join(args_strings)

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> Process:
        """Deserialize a dictionary into a process.

        Args:
            dictionary: A serialized proces object.

        Returns:
            The reconstructed process object.
        """
        dictionary = deepcopy(dictionary)
        command = dictionary.pop("command")
        command_args = shlex.split(command)
        dictionary["command_args"] = command_args
        process = cls(**dictionary)
        return process


class ExternalTestMixin(BaseTest):
    pass_code: ClassVar[str]
    # Class attributes
    is_external_test = True

    # Class constants
    STDOUT_PATH: ClassVar[Path]
    STDOUT_PATH = Path("std_out.txt")
    STDERR_PATH: ClassVar[Path]
    STDERR_PATH = Path("std_err.txt")
    EXITCODE_PATH: ClassVar[Path]
    EXITCODE_PATH = Path("exit_code.txt")

    def compute_status(self) -> TestStatus:
        """Compute the status of the test."""
        outputs = self._find_process_outputs()
        return self._interpret_process_outputs(outputs)

    @abstractmethod
    def generate_process(self) -> Process:
        """Generate the process that needs to be run."""

    @classmethod
    def _find_process_outputs(
        cls, search_dir: Optional[Path] = None
    ) -> dict[str, Path]:
        """Locate the output files from the executed process."""
        search_dir = search_dir or Path(".")
        outputs = {
            "std_out": search_dir / cls.STDOUT_PATH,
            "std_err": search_dir / cls.STDERR_PATH,
            "exit_code": search_dir / cls.EXITCODE_PATH,
        }

        for path in outputs.values():
            if not path.exists():
                message = f"Expected process output ({path}) does not exist."
                raise FileNotFoundError(message)
        return outputs

    def _interpret_process_outputs(self, outputs: dict[str, Path]) -> TestStatus:
        """Interpret the process output files to yield a test status."""
        exit_code = outputs["exit_code"].read_text()
        exit_code = exit_code.strip()
        if exit_code == self.pass_code:
            status = TestStatus.PASS
        else:
            status = TestStatus.FAIL
        return status

    # TODO: make changes to this package or the nextflow
    # workflow so that file mounting is handled cleaner
    @staticmethod
    def _short_string_path(path: Path, substring: str) -> str:
        # chech if substring is in path
        if substring not in path.as_posix():
            raise ValueError(f"{substring} not in {path}")
        # get index where staged folder is
        index = next(i for i, item in enumerate(path.parts) if substring in item)
        # shorten path starting from staged folder
        short_path = Path(*path.parts[index:])
        # wrap path string in quotes
        quote_path = str(short_path)
        return f"'{quote_path}'"

    # TODO: Include process in serialized test dictionary
    # def to_dict(self):
    #     dictionary = super(ExternalTestMixin, self).to_dict()
    #     process = self.generate_process()
    #     dictionary["process"] = process.to_dict()
    #     return dictionary


class InternalBaseTest(BaseTest):
    """Base class for all internal tests."""


class ExternalBaseTest(ExternalTestMixin, BaseTest):
    """Base class for all external tests."""



from dcqc.target import SingleTarget
from dcqc.tests.base_test import ExternalBaseTest, Process


class GrepDateTest(ExternalBaseTest):
    tier = 4
    pass_code = "1"
    target: SingleTarget

    def generate_process(self) -> Process:
        path = self.target.file.stage()
        string_path = self._short_string_path(path, "dcqc-staged-")

        command_args = [
            "grep",
            "-E",  # extended regular expression
            "-i",  # case insensitive
            "-a",  # treat input as text
            "-q",  # suppress output
            "'date|time'",  # match date or time
            string_path,
        ]
        process = Process(
            container="quay.io/biocontainers/coreutils:8.30--h14c3975_1000",
            command_args=command_args,
        )
        return process
